# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.text_helper

'''
Routines to help with text processing
'''
import re
import warnings
from itertools import zip_longest


# XXX: Do we want a case-insensitive separator regex flag?
def text_split(text: str, chunk_size: int, separator: str='\n\n', joiner=None, len_func=len):
    '''
    Split string and generate the sequence of chunks

    >>> from ogbujipt.text_helper import text_split
    >>> list(text_split('She sells seashells by the seashore', chunk_size=5, separator=' '))
    ['She', 'sells', 'seashells', 'by', 'the', 'seashore']
    >>> list(text_split('She sells seashells by the seashore', chunk_size=10, separator=' '))
    ['She sells', 'seashells', 'by the', 'seashore']
    # Notice case sensitivity, plus the fact that the separator is not included in the chunks
    >>> list(text_split('She sells seashells by the seashore', chunk_size=10, separator='s'))
    ['She ', 'ells ', 'ea', 'hell', ' by the ', 'ea', 'hore']
    >>> list(text_split('She\tsells seashells\tby the seashore', chunk_size=10, separator='\\s'))
    ['She\tsells', 'seashells', 'by the', 'seashore']
    >>> list(text_split('She\tsells seashells\tby the seashore', chunk_size=10, separator='\\s', joiner=' '))
    ['She sells', 'seashells', 'by the', 'seashore']

    Args:
        text (str): String to be split into chunks

        chunk_size (int): Guidance on maximum length (based on distance_function) of each chunk

        seperator (str, optional): Regex used to split `text` into sections. Do not include outer capturing parenthesis.
            Don't forget to use escaping where necessary.

        joiner (str, optional): Exact string used to rejoin any sections in order to meet target length
            defaults to using the literal match from the separator

        len_func (callable, optional): Function to measure chunk length, len() by default

    Returns:
        chunks (List[str]): List of chunks of the text provided
    '''
    assert separator, 'Separator must be non-empty'
    
    if ((not isinstance(text, str))
        or (not isinstance(separator, str))):
        raise ValueError(f'text and separator must be strings.\n'
                         f'Got {text.__class__} for text and {separator.__class__} for separator')
    
    if ((not isinstance(chunk_size, int)) or (chunk_size <= 0)):
        raise ValueError(f'chunk_size must be a positive integer, got {chunk_size}.')
    
    # Split the text by the separator
    if joiner is None:
        separator = f'({separator})'
    sep_pat = re.compile(separator)
    raw_split = re.split(sep_pat, text)

    # Rapid aid to understanding following logic:
    # data = ['a',' ','b','\t','c']
    # list(zip_longest(data[0::2], data[1::2], fillvalue=''))
    #     →[('a', ' '), ('b', '\t'), ('c', '')]
    fine_split = ([ i for i in zip_longest(raw_split[0::2], raw_split[1::2], fillvalue='') ]
                    if joiner is None else re.split(sep_pat, text))

    if len(fine_split) <= 1:
        warnings.warn(f'No splits detected. Perhaps a problem with separator? ({repr(separator)})?')

    curr_chunk = []
    chunk_len = 0

    for fs in fine_split:
        (fs, sep) = fs if joiner is None else (fs, joiner)
        if not fs: continue  # noqa E701
        sep_len = len_func(sep)
        len_fs = len_func(fs)
        # if len_fs > chunk_size:
        #     warnings.warn(f'One of the splits is larger than the chunk size. '
        #                   f'Consider increasing the chunk size or splitting the text differently.')

        if chunk_len + len_fs > chunk_size:
            chunk = ''.join(curr_chunk[:-1])
            if chunk: yield chunk  # noqa E701
            curr_chunk, chunk_len = [fs, sep], len_fs + sep_len
        else:
            curr_chunk.extend((fs, sep))
            chunk_len += len_fs + sep_len

    if curr_chunk:
        chunk = ''.join(curr_chunk[:-1])
        if chunk: yield chunk  # noqa E701


def text_split_fuzzy(text: str,
        chunk_size: int,
        chunk_overlap: int=0,
        separator: str='\n\n',
        joiner=None,
        len_func=len
    ):
    '''
    Split string in a "fuzzy" manner and generate the sequence of chunks
    Will generally not split the text into sequences of chunk_size length, but will instead preserve some overlap
    on either side of the given separators.
    This results in slightly larger chunks than the chunk_size number by itself suggests.

    >>> from ogbujipt.text_helper import text_split_fuzzy
    >>> chunks = list(text_split_fuzzy('she sells seashells by the seashore', chunk_size=5, separator=' '))

    Args:
        text (str): (Multiline) string to be split into chunks

        chunk_size (int): Target length (as determined by len_func) per chunk

        chunk_overlap (int, optional): Number of characters to overlap at the edges of chunks

        seperator (str, optional): Regex used to split `text` into sections. Do not include outer capturing parenthesis.
            Don't forget to use escaping where necessary.

        joiner (str, optional): Exact string used to rejoin any sections in order to meet target length
            defaults to using the literal match from the separator

        len_func (callable, optional): Function to measure chunk length, len() by default

    Yields:
        Sequence of chunks (str) from splitting
    '''
    assert separator, 'Separator must be non-empty'
    
    if ((not isinstance(text, str))
        or (not isinstance(separator, str))):
        raise ValueError(f'text and separator must be strings.\n'
                         f'Got {text.__class__} for text and {separator.__class__} for separator')
    
    if chunk_overlap == 0:
        msg = 'chunk_overlap must be a positive integer. For no overlap, use text_split() instead'
        raise ValueError(msg)

    if ((not isinstance(chunk_size, int))
        or (not isinstance(chunk_overlap, int))
        or (chunk_size <= 0)
        or (chunk_overlap < 0)
        or (chunk_size < chunk_overlap)):
        raise ValueError(f'chunk_size must be a positive integer, '
                         f'chunk_overlap must be a non-negative integer, and'
                         f'chunk_size must be greater than chunk_overlap.\n'
                         f'Got {chunk_size} chunk_size and {chunk_overlap} chunk_overlap.')

    # Split up the text by the separator
    # FIXME: Need a step for escaping regex
    sep_pat = re.compile(separator)
    fine_split = re.split(sep_pat, text)
    separator_len = len_func(separator)

    if len(fine_split) <= 1:
        warnings.warn(f'No splits detected. Perhaps a problem with separator? ({repr(separator)})?')

    at_least_one_chunk = False

    # Combine the small pieces into medium size chunks
    # chunks will accumulate processed text chunks as we go along
    # curr_chunk will be a list of subchunks comprising the main, current chunk
    # back_overlap will be appended, once ready, to the end of the previous chunk (if any)
    # fwd_overlap will be prepended, once ready, to the start of the next chunk
    prev_chunk = ''
    curr_chunk, curr_chunk_len = [], 0
    back_overlap, back_overlap_len = None, 0  # None signals not yet gathering
    fwd_overlap, fwd_overlap_len = None, 0

    for s in fine_split:
        if not s: continue  # noqa E701
        split_len = len_func(s) + separator_len
        # Check for full back_overlap (if relevant, i.e. back_overlap isn't None)
        if back_overlap is not None and (back_overlap_len + split_len > chunk_overlap):  # noqa: F821
            prev_chunk.extend(back_overlap)
            back_overlap, back_overlap_len = None, 0

        # Will adding this split take us into overlap room?
        if curr_chunk_len + split_len > (chunk_size - chunk_overlap):
            fwd_overlap, fwd_overlap_len = [], 0  # Start gathering

        # Will adding this split take us over chunk size?
        if curr_chunk_len + split_len > chunk_size:
            # If so, complete current chunk & start a new one

            # fwd_overlap should be non-None at this point, so check empty
            if not fwd_overlap and curr_chunk:
                # If empty, look back to make sure there is some overlap
                fwd_overlap.append(curr_chunk[-1])

            prev_chunk = separator.join(prev_chunk)
            if prev_chunk:
                at_least_one_chunk = True
                yield prev_chunk
            prev_chunk = curr_chunk
            # fwd_overlap intentionally not counted in running chunk length
            curr_chunk, curr_chunk_len = fwd_overlap, 0
            back_overlap, back_overlap_len = [], 0  # Start gathering
            fwd_overlap, fwd_overlap_len = None, 0  # Stop gathering

        if fwd_overlap is not None:
            fwd_overlap.append(s)
            fwd_overlap_len += split_len

        if back_overlap is not None:
            back_overlap.append(s)
            back_overlap_len += split_len

        curr_chunk.append(s)
        curr_chunk_len += split_len

    # Done with the splits; use the final back_overlap, if any
    if back_overlap:
        prev_chunk.extend(back_overlap)

    if at_least_one_chunk:
        yield separator.join(prev_chunk)
    else:
        # Degenerate case where no splits found & chunk size too large; just one big chunk
        yield text


def text_splitter(text: str,
        chunk_size: int,
        chunk_overlap: int=0,
        separator: str='\n\n',
        len_func=len
    ) -> list[str]:
    '''
    Split string into a sequence of chunks in a "fuzzy" manner. Will generally not split the text into sequences
    of chunk_size length, but will instead preserve some overlap on either side of the given separators.
    This results in slightly larger chunks than the chunk_size number by itself suggests.

    >>> from ogbujipt.text_helper import text_splitter
    >>> chunks = text_split_fuzzy('she sells seashells by the seashore', chunk_size=50, separator=' ')

    Args:
        text (str): (Multiline) String to be split into chunks

        chunk_size (int): Number of characters to include per chunk

        chunk_overlap (int, optional): Number of characters to overlap at the edges of chunks

        seperator (str, optional): String that already splits "text" into sections

        distance_function (callable, optional): Function to measure length, len() by default

    Returns:
        chunks (List[str]): List of chunks of the text provided
    '''
    warnings.warn('text_splitter() is deprecated. Use text_split_fuzzy() instead.')

    assert separator, 'Separator must be non-empty'
    
    if ((not isinstance(text, str))
        or (not isinstance(separator, str))):
        raise ValueError(f'text and separator must be strings.\n'
                         f'Got {text.__class__} for text and {separator.__class__} for separator')
    
    if chunk_overlap == 0:
        msg = 'chunk_overlap must be a positive integer. For no overlap, use text_split() instead'
        raise ValueError(msg)

    if ((not isinstance(chunk_size, int))
        or (not isinstance(chunk_overlap, int))
        or (chunk_size <= 0)
        or (chunk_overlap < 0)
        or (chunk_size < chunk_overlap)):
        raise ValueError(f'chunk_size must be a positive integer, '
                         f'chunk_overlap must be a non-negative integer, and'
                         f'chunk_size must be greater than chunk_overlap.\n'
                         f'Got {chunk_size} chunk_size and {chunk_overlap} chunk_overlap.')
    
    # Split up the text by the separator
    # FIXME: Need a step for escaping regex
    sep_pat = re.compile(separator)
    fine_split = re.split(sep_pat, text)
    separator_len = len_func(separator)

    if len(fine_split) <= 1:
        warnings.warn(f'No splits detected. Perhaps a problem with separator? ({repr(separator)})?')

    # Combine the small pieces into medium size chunks
    # chunks will accumulate processed text chunks as we go along
    # curr_chunk will be a list of subchunks comprising the main, current chunk
    # back_overlap will be appended, once ready, to the end of the previous chunk (if any)
    # fwd_overlap will be prepended, once ready, to the start of the next chunk
    chunks = []
    curr_chunk, curr_chunk_len = [], 0
    back_overlap, back_overlap_len = None, 0  # None signals not yet gathering
    fwd_overlap, fwd_overlap_len = None, 0

    for s in fine_split:
        if not s: continue  # noqa E701
        split_len = len_func(s) + separator_len
        # Check for full back_overlap (if relevant, i.e. back_overlap isn't None)
        if back_overlap is not None and (back_overlap_len + split_len > chunk_overlap):  # noqa: F821
            chunks[-1].extend(back_overlap)
            back_overlap, back_overlap_len = None, 0

        # Will adding this split take us into overlap room?
        if curr_chunk_len + split_len > (chunk_size - chunk_overlap):
            fwd_overlap, fwd_overlap_len = [], 0  # Start gathering

        # Will adding this split take us over chunk size?
        if curr_chunk_len + split_len > chunk_size:
            # If so, complete current chunk & start a new one

            # fwd_overlap should be non-None at this point, so check empty
            if not fwd_overlap and curr_chunk:
                # If empty, look back to make sure there is some overlap
                fwd_overlap.append(curr_chunk[-1])

            chunks.append(curr_chunk)
            # fwd_overlap intentionally not counted in running chunk length
            curr_chunk, curr_chunk_len = fwd_overlap, 0
            back_overlap, back_overlap_len = [], 0  # Start gathering
            fwd_overlap, fwd_overlap_len = None, 0  # Stop gathering

        if fwd_overlap is not None:
            fwd_overlap.append(s)
            fwd_overlap_len += split_len

        if back_overlap is not None:
            back_overlap.append(s)
            back_overlap_len += split_len

        curr_chunk.append(s)
        curr_chunk_len += split_len

    # Done with the splits; use the final back_overlap, if any
    if back_overlap:
        chunks[-1].extend(back_overlap)

    # Concatenate all the split parts of all the chunks
    chunks = [separator.join(c) for c in chunks]

    # Handle degenerate case where no splits found & chunk size too large
    # Just becomes one big chunk
    if not chunks:
        chunks = [text]

    # chunks.append(separator.join(curr_chunk))
    return chunks



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/embedding/test_pgvector_data.py
'''
After setup as described in the README.md for this directory, run the tests with:

pytest test

or, for just this test module:

`pytest test/embedding/test_pgvector_data.py`

Uses fixtures from test/embedding/conftest.py in current & parent directories, including Postgres database connection
'''

import pytest
from unittest.mock import MagicMock, DEFAULT  # noqa: F401

import numpy as np

from ogbujipt.embedding.pgvector import match_exact, match_oneof


KG_STATEMENTS = [  # Demo data
    ("👤 Alikiba `releases` 💿 'Yalaiti'",
        {'url': 'https://njok.com/yalaiti-lyrics/', 'primary': True, 'when': '2023-11-29'}),
    ("👤 Sabah Salum `featured_in` 💿 'Yalaiti'",
        {'url': 'https://njok.com/yalaiti-lyrics/', 'primary': False, 'when': '2023-11-29'}),
    ('👤 Kukbeatz `collab_with` 👤 Ruger',
        {'url': 'https://njok.com/all-of-us-lyrics/', 'primary': True, 'when': '2023-11-25'}),
    ('💿 All of Us `song_by` 👤 Kukbeatz & Ruger',
        {'url': 'https://njok.com/all-of-us-lyrics/', 'primary': False, 'when': '2023-11-25'}),
    ('👤 Blaqbonez `collab_with` 👤 Fireboy DML',
        {'url': 'https://njok.com/fireboy-dml-collab/', 'primary': True, 'when': '2023-11-19'})
]


# XXX: Move to a fixture?
# Definitely don't want to even import SentenceTransformer class due to massive side-effects
class SentenceTransformer(object):
    def __init__(self, model_name_or_path):
        self.encode = MagicMock()


@pytest.mark.asyncio
async def test_insert_data_vector(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])

    item1_text = KG_STATEMENTS[0][0]
    item1_meta = KG_STATEMENTS[0][1]

    # Insert data
    for index, (text, meta) in enumerate(KG_STATEMENTS):
        await DB.insert(    # Insert the row into the table
            content=text,   # text to be embedded
            metadata=meta,  # Tag metadata
        )

    assert await DB.count_items() == len(KG_STATEMENTS), Exception('Incorrect number of documents after insertion')

    # search table with perfect match
    result = await DB.search(text=item1_text, limit=3)
    # assert result is not None, Exception('No results returned from perfect search')

    # Even though the embedding is mocked, the stored text should be faithful
    row = next(result)
    assert row.content == item1_text, 'text mismatch'
    assert row.metadata == item1_meta, 'Metadata mismatch'


@pytest.mark.asyncio
async def test_insertmany_data_vector(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])

    item1_text = KG_STATEMENTS[0][0]
    item1_meta = KG_STATEMENTS[0][1]

    # Insert data using insert_many()
    # dataset = ((text, metadata) for (text, metadata) in KG_STATEMENTS)
    
    await DB.insert_many(KG_STATEMENTS)

    assert await DB.count_items() == len(KG_STATEMENTS), Exception('Incorrect number of documents after insertion')

    # search table with perfect match
    result = await DB.search(text=item1_text, limit=3)
    # assert result is not None, Exception('No results returned from perfect search')

    # Even though the embedding is mocked, the stored text should be faithful
    row = next(result)
    assert row.content == item1_text
    
    # Adjusted assertion for metadata comparison
    assert row.metadata == item1_meta, "Metadata mismatch"


@pytest.mark.asyncio
async def test_search_with_filter(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])

    # item1_text = KG_STATEMENTS[0][0]

    # Insert data using insert_many()
    # dataset = ((text, metadata) for (text, metadata) in KG_STATEMENTS)

    await DB.insert_many(KG_STATEMENTS)

    # search table with perfect match, but only where primary is set to True
    primary_filt = match_exact('primary', True)
    result = list(await DB.search(text='Kukbeatz and Ruger', meta_filter=primary_filt))
    assert len(result) == 3

    result = list(await DB.search(text='Kukbeatz and Ruger', meta_filter=primary_filt, limit=1))
    assert len(result) == 1


@pytest.mark.asyncio
async def test_search_with_date_filter(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])

    # item1_text = KG_STATEMENTS[0][0]

    # Insert data using insert_many()
    # dataset = ((text, metadata) for (text, metadata) in KG_STATEMENTS)

    await DB.insert_many(KG_STATEMENTS)

    # search table with perfect match, but only where primary is set to True
    primary_filt = match_exact('when', '2023-11-29')
    result = list(await DB.search(text='Kukbeatz and Ruger', meta_filter=primary_filt))
    assert len(result) == 2

    primary_filt = match_exact('when', '2023-11-19')
    result = list(await DB.search(text='Kukbeatz and Ruger', meta_filter=primary_filt))
    assert len(result) == 1


@pytest.mark.asyncio
async def test_search_with_date_filter_match_oneof(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])

    # item1_text = KG_STATEMENTS[0][0]

    # Insert data using insert_many()
    # dataset = ((text, metadata) for (text, metadata) in KG_STATEMENTS)

    await DB.insert_many(KG_STATEMENTS)

    # search table with perfect match, but only where primary is set to True
    primary_filt = match_oneof('when', ('2023-11-29',))
    result = list(await DB.search(text='Kukbeatz and Ruger', meta_filter=primary_filt))
    assert len(result) == 2

    primary_filt = match_oneof('when', ('2023-11-29', '2023-11-19'))
    result = list(await DB.search(text='Kukbeatz and Ruger', meta_filter=primary_filt))
    assert len(result) == 3

    primary_filt = match_oneof('when', ('2023-11-29', '2023-11-25', '2023-11-19'))
    result = list(await DB.search(text='Kukbeatz and Ruger', meta_filter=primary_filt))
    assert len(result) == 5


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.embedding.pgvector_data_doc

'''
Vector databases embeddings using PGVector
'''

from typing import Iterable, Callable, List, Sequence

from ogbujipt.embedding.pgvector import PGVectorHelper, asyncpg, process_search_response

__all__ = ['DataDB']

# ------ SQL queries ---------------------------------------------------------------------------------------------------
# PG only supports proper query arguments (e.g. $1, $2, etc.) for values, not for table or column names
# Table names are checked to be legit sequel table names, and embed_dimension is assured to be an integer

CREATE_TABLE_BASE = '''-- Create a table to hold embedded documents or data
CREATE TABLE IF NOT EXISTS {{table_name}} (
    id BIGSERIAL PRIMARY KEY,
    embedding VECTOR({{embed_dimension}}),  -- embedding vectors (array dimension)
    content TEXT NOT NULL,                  -- text content of the chunk
    metadata JSON                           -- additional metadata of the chunk
{extra_fields});
'''

# CREATE_DOC_TABLE = CREATE_TABLE_BASE.format(extra_fields='''\
# ,    title TEXT,                            -- title of file
#     page_numbers INTEGER[]                  -- page number of the document that the chunk is found in
# ''')

CREATE_DATA_TABLE = CREATE_TABLE_BASE.format(extra_fields='')

INSERT_BASE = '''-- Insert a document into a table
INSERT INTO {{table_name}} (
    embedding,
    content,
    metadata
    {extra_fields}) VALUES ($1, $2, $3{extra_vals});
'''

INSERT_DATA = INSERT_BASE.format(extra_fields='', extra_vals='')

QUERY_TABLE_BASE = '''-- Semantic search a document
SELECT * FROM (  -- Subquery to calculate cosine similarity, required to use the alias in the WHERE clause
    SELECT
        1 - (embedding <=> $1) AS cosine_similarity,
        content,
        metadata
        {extra_fields}
    FROM
        {{table_name}}
) subquery
{{where_clauses}}
ORDER BY
    cosine_similarity DESC
{{limit_clause}};
'''

# QUERY_DOC_TABLE = QUERY_TABLE_BASE.format(extra_fields='''\
# ,        title
# ''')

QUERY_DATA_TABLE = QUERY_TABLE_BASE.format(extra_fields='')

# TITLE_WHERE_CLAUSE = 'title = {query_title}  -- Equals operator\n'

# PAGE_NUMBERS_WHERE_CLAUSE = 'page_numbers && {query_page_numbers}  -- Overlap operator \n'

THRESHOLD_WHERE_CLAUSE = 'cosine_similarity >= {query_threshold}\n'
# ------ SQL queries ---------------------------------------------------------------------------------------------------


class DataDB(PGVectorHelper):
    ''' Specialize PGvectorHelper for data (snippets) '''
    async def create_table(self) -> None:
        '''
        Create the table to hold embedded documents
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    CREATE_DATA_TABLE.format(
                        table_name=self.table_name,
                        embed_dimension=self._embed_dimension)
                    )
    
    async def insert(
            self,
            content: str,
            metadata: dict | None = None
    ) -> None:
        '''
        Update a table with one embedded document

        Args:
            content (str): text content of the document

            metadata (dict, optional): additional metadata of the chunk
        '''
        # Get the embedding of the content as a PGvector compatible list
        content_embedding = self._embedding_model.encode(content)

        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    INSERT_DATA.format(table_name=self.table_name),
                    content_embedding,
                    content,
                    metadata
                )

    async def insert_many(
            self,
            content_list: Iterable[tuple[str, dict]]
    ) -> None:
        '''
        Update a table with one or (presumably) more embedded documents

        Semantically equivalent to multiple insert calls, but uses executemany for efficiency

        Args:
            content_list: List of tuples, each of the form: (content, metadata), where
                `content` is a string and
                `metadata` is a dictionary
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.executemany(
                    INSERT_DATA.format(table_name=self.table_name),
                    (
                        # Does this need to be .tolist()?
                        (self._embedding_model.encode(content), content, metadata)
                        for content, metadata in content_list
                    )
                )

    async def search(
            self,
            text: str,
            threshold: float | None = None,
            meta_filter: Callable | List[Callable] | None = None,
            limit: int = 0
    ) -> list[asyncpg.Record]:
        '''
        Similarity search documents using a query string

        Args:
            text (str): string to compare against items in the table.
                This will be a vector/fuzzy/nearest-neighbor type search.

            threshold: minimum vector similarity to return

            meta_filter: specifies further restrictions on results, via one or more functions setting up SQL conditions
                for the metadata field. Can be a callable or list of callables, which will be ANDed together
                e.g.
                ```
                from ogbujipt.embedding.pgvector import match_exact
                page_one_filt = match_exact('page', 1)
                await DB.search(text='Hello!', meta_filter=page_one_filt)
                ```
                Provided filter helpers such as match_exact provide some SQL injection safety. If you roll your own,
                be careful of injection.
            
            limit (int, optional): maximum number of results to return (useful for top-k query)
                Default is no limit

        Returns:
            generator which yields the rows os the query results ass attributable dicts
        '''
        if threshold is not None:
            if not isinstance(threshold, float) or (threshold < 0) or (threshold > 1):
                raise TypeError('threshold must be a float between 0.0 and 1.0')
        if not isinstance(limit, int):
            raise TypeError('limit must be an integer')  # Guard against injection
        meta_filter = meta_filter or ()
        if not isinstance(meta_filter, Sequence):
            meta_filter = (meta_filter,)

        # Get the embedding of the query string as a PGvector compatible list
        query_embedding = list(self._embedding_model.encode(text))

        # Build where clauses
        query_args = [query_embedding]
        where_clauses = []
        if threshold is not None:
            query_args.append(threshold)
            where_clauses.append(THRESHOLD_WHERE_CLAUSE.format(query_threshold=f'${len(query_args)}'))

        for mf in meta_filter:
            assert callable(mf), 'All meta_filter items must be callable'
            clause, pval = mf()
            query_args.append(pval)
            where_clauses.append(clause.format(len(query_args)))

        where_clauses_str = 'WHERE\n' + 'AND\n'.join(where_clauses) if where_clauses else ''

        if limit:
            limit_clause = f'LIMIT {limit}\n'
        else:
            limit_clause = ''

        # print(QUERY_DATA_TABLE.format(table_name=self.table_name, where_clauses=where_clauses_str,
        #                             limit_clause=limit_clause,
        #     ))
        # print(query_args)

        # Execute the search via SQL
        async with self.pool.acquire() as conn:
            # Uncomment to debug
            # from asyncpg import utils
            # print(await utils._mogrify(
            #     conn,
            #     QUERY_DATA_TABLE.format(table_name=self.table_name, where_clauses=where_clauses_str,
            #                             limit_clause=limit_clause,
            #     ),
            #     query_args
            # ))
            search_results = await conn.fetch(
                QUERY_DATA_TABLE.format(table_name=self.table_name, where_clauses=where_clauses_str,
                                        limit_clause=limit_clause,
                ),
                *query_args
            )
        return process_search_response(search_results)



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt/demo/chat_web_selects.py
'''
Advanced, "Chat my docs" demo, using docs from the web

Download one or more web pages and query an LLM using them as context.
Works especially well with airoboros self-hosted LLM.

Vector store: Qdrant - https://qdrant.tech/
    Alternatives: pgvector & Chroma (built-in support from OgbujiPT), Faiss, Weaviate, etc.
Text to vector (embedding) model: 
    Alternatives: https://www.sbert.net/docs/pretrained_models.html / OpenAI ada002

Needs access to an OpenAI-like service. Default assumption is self-hosted
via e.g. llama-cpp-python or text-generation-webui

Assume for the following it's at host localhost, port 8000.
MAKE SURE YOU USE A WORKING SERVER BECAUSE THIS IS A DEMO & THE ERROR HANDLING IS SIMPLISTIC

pip install prerequisites, in addition to OgbujiPT cloned dir:

click sentence_transformers qdrant-client httpx html2text amara3.xml

```sh
python demo/chat_web_selects.py --apibase http://localhost:8000 "www.newworldencyclopedia.org/entry/Igbo_People"
```

An example question might be "Who are the neighbors of the Igbo people?"

You can tweak it with the following command line options:
--verbose - print more information while processing (for debugging)
--limit (max number of chunks to retrieve for use as context)
--chunk-size (characters per chunk, while prepping to create embeddings)
--chunk-overlap (character overlap between chunks, while prepping to create embeddings)
--question (The user question; if None (the default), prompt the user interactively)
'''
# en.wikipedia.org/wiki/Igbo_people|ahiajoku.igbonet.com/2000/|en.wikivoyage.org/wiki/Igbo_phrasebook"
import asyncio
import os

import click
import httpx
import html2text

from ogbujipt.llm_wrapper import openai_chat_api, prompt_to_chat
from ogbujipt.text_helper import text_split_fuzzy
from ogbujipt.embedding.qdrant import collection


# Avoid re-entrace complaints from huggingface/tokenizers
os.environ['TOKENIZERS_PARALLELISM'] = 'false'

# default https://huggingface.co/sentence-transformers/all-MiniLM-L6-v2
DOC_EMBEDDINGS_LLM = 'all-MiniLM-L12-v2'

COLLECTION_NAME = 'chat-web-selects'
USER_PROMPT = 'What do you want to know from the site(s)?\n'

DOTS_SPACING = 0.2  # Number of seconds between each dot printed to console


async def indicate_progress(pause=DOTS_SPACING):
    while True:
        print('.', end='', flush=True)
        await asyncio.sleep(pause)


async def read_site(url, coll, chunk_size, chunk_overlap):
    # Crude check; good enough for demo
    if not url.startswith('http'): url = 'https://' + url  # noqa E701
    print('Downloading & processing', url)
    async with httpx.AsyncClient(verify=False) as client:
        resp = await client.get(url)
        html = resp.content.decode(resp.encoding or 'utf-8')

    text = html2text.html2text(html)

    # Split text into chunks
    chunks = text_split_fuzzy(text, chunk_size=chunk_size, chunk_overlap=chunk_overlap, separator='\n')

    # print('\n\n'.join([ch[:100] for ch in chunks]))
    # Crude—for demo. Set URL metadata for all chunks to doc URL
    metas = [{'url': url}]*len(chunks)
    # Add the text to the collection
    coll.update(texts=chunks, metas=metas)
    print(f'{coll.count()} chunks added to collection')


async def async_main(oapi, sites, verbose, limit, chunk_size, chunk_overlap, question):
    # Automatic download of embedding model from HuggingFace
    # Seem to be reentrancy issues with HuggingFace; defer import
    from sentence_transformers import SentenceTransformer
    embedding_model = SentenceTransformer(DOC_EMBEDDINGS_LLM)
    # Sites fuel in-memory Qdrant vector DB instance
    coll = collection(COLLECTION_NAME, embedding_model)

    # Download & process sites in parallel, loading their content & metadata into a knowledgebase
    url_task_group = asyncio.gather(*[
        asyncio.create_task(read_site(site, coll, chunk_size, chunk_overlap)) for site in sites.split('|')])
    indicator_task = asyncio.create_task(indicate_progress())
    tasks = [indicator_task, url_task_group]
    done, _ = await asyncio.wait(
        tasks, return_when=asyncio.FIRST_COMPLETED)

    # Main chat loop
    done = False
    while not done:
        print()
        if question:
            user_question = question
        else:
            user_question = input(USER_PROMPT)
        if user_question.strip() == 'done':
            break

        docs = coll.search(user_question, limit=limit)
        if verbose:
            print(docs)
        if docs:
            # Collects "chunked_doc" into "gathered_chunks"
            gathered_chunks = '\n\n'.join(
                doc.payload['_text'] for doc in docs if doc.payload)

            # Build system message with the doc chunks as provided context
            # In practice we'd use word loom to load the propts, as demoed in multiprocess.py
            sys_prompt = '''\
You are a helpful assistant, who answers questions directly and as briefly as possible.
Consider the following context and answer the user\'s question.
If you cannot answer with the given context, just say so.\n\n'''
            sys_prompt += gathered_chunks + '\n\n'
            messages = prompt_to_chat(user_question, system=sys_prompt)
            if verbose:
                print('-'*80, '\n', messages, '\n', '-'*80)

            # The rest is much like in demo/alpaca_multitask_fix_xml.py
            model_params = dict(
                max_tokens=1024,  # Limit number of generated tokens
                top_p=1,  # AKA nucleus sampling; can increase generated text diversity
                frequency_penalty=0,  # Favor more or less frequent tokens
                presence_penalty=1,  # Prefer new, previously unused tokens
                temperature=0.1)

            indicator_task = asyncio.create_task(indicate_progress())
            llm_task = asyncio.Task(oapi(messages, **model_params))
            tasks = [indicator_task, llm_task]
            done, _ = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED)

            # proper cleanup of indicator task, which will still be pending/running
            indicator_task.cancel()

            # Instance of openai.openai_object.OpenAIObject, with lots of useful info
            retval = next(iter(done)).result()
            if verbose:
                print(type(retval))
            # Response is a json-like object; extract the text
            if verbose:
                print('\nFull response data from LLM:\n', retval)

            # just get back the text of the response
            print('\nResponse text from LLM:\n\n', retval.first_choice_text)


# Command line arguments defined in click decorators
@click.command()
@click.option('--verbose/--no-verbose', default=False)
@click.option('--chunk-size', type=int, default=200,
              help='Number of characters to include per chunk')
@click.option('--chunk-overlap', type=int, default=20,
              help='Number of characters to overlap at the edges of chunks')
@click.option('--limit', default=4, type=int,
              help='Maximum number of chunks matched against the posed question to use as context for the LLM')
@click.option('--openai-key',
              help='OpenAI API key. Leave blank to specify self-hosted model via --host & --port')
@click.option('--apibase', default='http://127.0.0.1:8000', help='OpenAI API base URL')
@click.option('--model', default='', type=str, 
              help='OpenAI model to use (see https://platform.openai.com/docs/models).'
              'Use only with --openai-key')
@click.option('--question', default=None, help='The question to ask (or prompt for one)')
@click.argument('sites')
def main(verbose, chunk_size, chunk_overlap, limit, openai_key, apibase, model, question, sites):
    # Use OpenAI API if specified, otherwise emulate with supplied URL info
    if openai_key:
        oapi = openai_chat_api(api_key=openai_key, model=(model or 'gpt-3.5-turbo'))
    else:
        oapi = openai_chat_api(model=model, base_url=apibase)

    asyncio.run(async_main(oapi, sites, verbose, limit, chunk_size, chunk_overlap, question))


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_text_splitter.py
'''
pytest test

or

pytest test/test_text_splitter.py
'''
# ruff: noqa: E501

import pytest

from ogbujipt.text_helper import text_split, text_split_fuzzy

# Also test: chunk_overlap > chunk_size, chunk_overlap == chunk_size, etc.


def test_split_basic(BASIC_EVEN_BLOCK):
    # Need to test chunk_size > total length
    # chunks = text_split_fuzzy(BASIC_EVEN_BLOCK, chunk_size=100, chunk_overlap=10)
    chunks = list(text_split_fuzzy(
        BASIC_EVEN_BLOCK, chunk_size=21, chunk_overlap=3, separator='\n'))
    # import pprint; pprint.pprint(chunks)  # noqa
    assert len(chunks) == 3
    assert chunks == [
        '0123456789\nabcdefghij', '0123456789\nabcdefghij\nABCDEFGHIJ',
        'abcdefghij\nABCDEFGHIJ\nklmnopqrst'
        ]


# @pytest.mark.skip(reason="Still working on it")
def test_split_poem(COME_THUNDER_POEM):
    # Just use a subset of the poem
    poem_text = COME_THUNDER_POEM.partition('The arrows of God')[0]
    chunks = list(text_split_fuzzy(poem_text, chunk_size=200, chunk_overlap=30))
    # print(chunks[0], '\n---')
    assert chunks[0] == 'Now that the triumphant march has entered the last street corners,\nRemember, O dancers, the thunder among the clouds…\n\nNow that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…'
    assert chunks[1] == 'Now that the triumphant march has entered the last street corners,\nRemember, O dancers, the thunder among the clouds…\n\nNow that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…\n\nThe smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.'
    assert chunks[2] == 'Now that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…\n\nThe smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.\n\nThe drowsy heads of the pods in barren farmlands witness it,\nThe homesteads abandoned in this century’s brush fire witness it:\nThe myriad eyes of deserted corn cobs in burning barns witness it:\nMagic birds with the miracle of lightning flash on their feathers…'
    # import pprint; pprint.pprint(chunks)  # noqa
    assert len(chunks) == 3


def test_separator_not_found(BASIC_EVEN_BLOCK):
    with pytest.warns(UserWarning):
        chunks = list(text_split_fuzzy(
            BASIC_EVEN_BLOCK, chunk_size=21, chunk_overlap=3, separator='!!!'))
    assert len(chunks) == 1
    assert len(chunks[0]) == len(BASIC_EVEN_BLOCK)


def test_separator_chunk_size_too_large(BASIC_EVEN_BLOCK):
    # Just use a subset of the poem
    with pytest.warns(UserWarning):
        chunks = list(text_split_fuzzy(
            BASIC_EVEN_BLOCK, chunk_size=100, chunk_overlap=10, separator='!!!'))
    assert len(chunks) == 1
    assert len(chunks[0]) == len(BASIC_EVEN_BLOCK)

def test_zero_overlap(LOREM_IPSUM):
    chunks = list(text_split(LOREM_IPSUM, chunk_size=100, separator=' '))
    assert len(chunks) == 10
    assert chunks[0] == 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Fusce vestibulum nisi eget mauris'
    assert chunks[2] == 'massa, in varius nulla ex vel ipsum. Nullam vitae eros nec ante sagittis luctus. Nullam scelerisque'
    assert chunks[3] == 'dolor eu orci iaculis, at convallis nulla luctus. Praesent eget ex id arcu facilisis varius vel id'
    assert chunks[-1] == 'quam justo at elit.'
    for chunk in chunks:
        assert len(chunk) <= 100


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.embedding.pgvector_message

'''
Vector embeddings DB feature for messaging (chat, etc.) using PGVector
'''

from uuid import UUID
from datetime import datetime, timezone
from typing import Iterable

from ogbujipt.config import attr_dict
from ogbujipt.embedding.pgvector import PGVectorHelper, asyncpg, process_search_response

__all__ = ['MessageDB']

# ------ SQL queries ---------------------------------------------------------------------------------------------------
# PG only supports proper query arguments (e.g. $1, $2, etc.) for values, not for table or column names
# Table names are checked to be legit sequel table names, and embed_dimension is assured to be an integer

CREATE_MESSAGE_TABLE = '''-- Create a table to hold individual messages (e.g. from a chatlog) and their metadata
CREATE TABLE IF NOT EXISTS {table_name} (
    ts TIMESTAMP WITH TIME ZONE,              -- timestamp of the message
    history_key UUID,                         -- uunique identifier for contextual message history
    role TEXT,                                -- role of the message (meta ID such as 'system' or user,
                                              -- or an ID associated with the sender)
    content TEXT NOT NULL,                    -- text content of the message
    embedding VECTOR({embed_dimension}),      -- embedding vectors (array dimension)
    metadata JSON                             -- additional metadata of the message
);
'''

INSERT_MESSAGE = '''-- Insert a message into a chatlog
INSERT INTO {table_name} (
    history_key,
    role,
    content,
    embedding,
    ts,
    metadata
) VALUES ($1, $2, $3, $4, $5, $6);
'''

CLEAR_MESSAGE = '''-- Deletes matching messages
DELETE FROM {table_name}
WHERE
    history_key = $1
'''

RETURN_MESSAGES_BY_HISTORY_KEY = '''-- Get entire chatlog by history key
SELECT
    ts,
    role,
    content,
    metadata
FROM
    {table_name}
WHERE
    history_key = $1
{since_clause}
ORDER BY
    ts DESC
{limit_clause};
'''

SEMANTIC_QUERY_MESSAGE_TABLE = '''-- Find messages with closest semantic similarity
SELECT
    1 - (embedding <=> $1) AS cosine_similarity,
    ts,
    role,
    content,
    metadata
FROM {table_name}
{where_clauses}
ORDER BY
    cosine_similarity DESC
{limit_clause};
'''

# The cosine_similarity alias is not available in the WHERE clause, so use a nested SELECT
SEMANTIC_QUERY_MESSAGE_TABLE = '''-- Find messages with closest semantic similarity
SELECT
    cosine_similarity,
    ts,
    role,
    content,
    metadata
FROM
    (SELECT
        history_key,
        1 - (embedding <=> $1) AS cosine_similarity,
        ts,
        role,
        content,
        metadata FROM {table_name}) AS main
{where_clauses}
{limit_clause};
'''

THRESHOLD_WHERE_CLAUSE = 'main.cosine_similarity >= {query_threshold}\n'

DELETE_OLDEST_MESSAGES = '''-- Delete oldest messages for given history key, such that only the newest N messages remain
DELETE FROM {table_name} t_outer
WHERE
    t_outer.history_key = $1
AND
    t_outer.ctid NOT IN (
        SELECT t_inner.ctid
        FROM {table_name} t_inner
        WHERE
            t_inner.history_key = $1
        ORDER BY
            t_inner.ts DESC
        LIMIT $2
);
'''

# Delete after full comfort with windowed implementation
# TEMPQ = '''
#         SELECT t_inner.ctid
#         FROM {table_name} t_inner
#         WHERE
#             t_inner.history_key = $1
#         ORDER BY
#             t_inner.ts DESC
#         LIMIT $2;
# '''

# ------ Class implementations ---------------------------------------------------------------------------------------

class MessageDB(PGVectorHelper):
    def __init__(self, embedding_model, table_name: str, pool: asyncpg.pool.Pool, window=0):
        '''
        Helper class for messages/chatlog storage and retrieval

        Args:
            embedding (SentenceTransformer): SentenceTransformer object of your choice
            https://huggingface.co/sentence-transformers
            window (int, optional): number of messages to maintain in the DB. Default is 0 (all messages)
        '''
        super().__init__(embedding_model, table_name, pool)
        self.window = window

    @classmethod
    async def from_conn_params(cls, embedding_model, table_name, host, port, db_name, user, password, window=0) -> 'MessageDB': # noqa: E501
        obj = await super().from_conn_params(embedding_model, table_name, host, port, db_name, user, password)
        obj.window = window
        return obj

    ''' Specialize PGvectorHelper for messages, e.g. chatlogs '''
    async def create_table(self):
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    CREATE_MESSAGE_TABLE.format(
                        table_name=self.table_name,
                        embed_dimension=self._embed_dimension
                    )
                )

    async def insert(
            self,
            history_key: UUID,
            role: str,
            content: str,
            # Timestamp later in order than SQL might suggest because of the defaulting
            timestamp: datetime | None = None,
            metadata: dict | None = None
    ) -> None:
        '''
        Update a table with one message's embedding

        Args:
            history_key (str): history key (unique identifier) of the chatlog this message belongs to

            role (str): role of the message (e.g. system, user, assistant, tool)

            content (str): text content of the message

            timestamp (datetime, optional): timestamp of the message, defaults to current time

            metadata (dict, optional): additional metadata of the message

        Returns:
            generator which yields the rows os the query results ass attributable dicts
        '''
        if not timestamp:
            timestamp = datetime.utcnow().replace(tzinfo=timezone.utc)

        # Get the embedding of the content as a PGvector compatible list
        content_embedding = self._embedding_model.encode(content)

        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    INSERT_MESSAGE.format(table_name=self.table_name),
                    history_key,
                    role,
                    content,
                    content_embedding.tolist(),
                    timestamp,
                    metadata
                )
            # print(f'{self.window=}, Pre-count: {await self.count_items()}')
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                if self.window:
                    await conn.execute(
                        DELETE_OLDEST_MESSAGES.format(table_name=self.table_name),
                        history_key,
                        self.window)
        # async with self.pool.acquire() as conn:
        #     print(f'{self.window=}, Post-count: {await self.count_items()}, {list(await conn.fetch(TEMPQ.format(table_name=self.table_name), history_key, self.window))}')  # noqa E501

    async def insert_many(
            self,
            content_list: Iterable[tuple[str, list[str]]]
    ) -> None:
        '''
        Update table with one or (presumably) more message embedding

        Semantically equivalent to multiple insert calls, but uses executemany for efficiency

        Args:
            content_list: List of tuples, each of the form: (history_key, role, text, timestamp, metadata)
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.executemany(
                    INSERT_MESSAGE.format(table_name=self.table_name),
                    (
                        (hk, role, text, self._embedding_model.encode(text), ts, metadata)
                        for hk, role, text, ts, metadata in content_list
                    )
                )
            # print(f'{self.window=}, Pre-count: {await self.count_items()}')  # noqa E501
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                if self.window:
                    # Set uniquifies the history keys
                    for hk in {hk for hk, _, _, _, _ in content_list}:
                        await conn.execute(
                            DELETE_OLDEST_MESSAGES.format(table_name=self.table_name),
                            hk, self.window)
        # async with self.pool.acquire() as conn:
        #     print(f'{self.window=}, {hk=}, Post-count: {await self.count_items()}, {list(await conn.fetch(TEMPQ.format(table_name=self.table_name), hk, self.window))}')  # noqa E501

    async def clear(
            self,
            history_key: UUID
    ) -> None:
        '''
        Remove all entries in a message history entries

        Args:
            history_key (str): history key (unique identifier) to match
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    CLEAR_MESSAGE.format(
                        table_name=self.table_name,
                    ),
                    history_key
                )

    # XXX: Change to a generator
    async def get_messages(
            self,
            history_key: UUID | str,
            since: datetime | None = None,
            limit: int = 0
    ): # -> list[asyncpg.Record]:
        '''
        Retrieve entries in a message history

        Args:
            history_key (str): history key (unique identifier) to match; string or object
            since (datetime, optional): only return messages after this timestamp
            limit (int, optional): maximum number of messages to return. Default is all messages
        Returns:
            generates asyncpg.Record instances of resulting messages
        '''
        if not isinstance(history_key, UUID):
            history_key = UUID(history_key)
            # msg = f'history_key must be a UUID, not {type(history_key)} ({history_key}))'
            # raise TypeError(msg)
        if not isinstance(since, datetime) and since is not None:
            msg = 'since must be a datetime or None'
            raise TypeError(msg)
        if not isinstance(limit, int):
            raise TypeError('limit must be an integer')  # Guard against injection

        qparams = [history_key]
        # Build query
        if since:
            # Don't really need the ${len(qparams) + N} thing here (first optional), but used for consistency
            since_clause = f' AND ts > ${len(qparams) + 1}'
            qparams.append(since)
        else:
            since_clause = ''
        if limit:
            limit_clause = f'LIMIT ${len(qparams) + 1}'
            qparams.append(limit)
        else:
            limit_clause = ''

        # Execute
        async with self.pool.acquire() as conn:
            message_records = await conn.fetch(
                RETURN_MESSAGES_BY_HISTORY_KEY.format(
                    table_name=self.table_name,
                    since_clause=since_clause,
                    limit_clause=limit_clause
                ),
                *qparams
            )

        return (attr_dict({
                'ts': record['ts'],
                'role': record['role'],
                'content': record['content'],
                'metadata': record['metadata']
            }) for record in message_records)

    async def search(
            self,
            history_key: UUID,
            text: str,
            since: datetime | None = None,
            threshold: float | None = None,
            limit: int = 1
    ) -> list[asyncpg.Record]:
        '''
        Similarity search documents using a query string

        Args:
            history_key (str): history key for the conversation to query
            text (str): string to compare against items in the table
            since (datetime, optional): only return results after this timestamp
            limit (int, optional): maximum number of messages to return; for top-k type query. Default is 1
        Returns:
            list[asyncpg.Record]: list of search results
                (asyncpg.Record objects are similar to dicts, but allow for attribute-style access)
        '''
        # Type checks
        if threshold is not None:
            if not isinstance(threshold, float) or (threshold < 0) or (threshold > 1):
                raise TypeError('threshold must be a float between 0.0 and 1.0')
        if not isinstance(limit, int):
            raise TypeError('limit must be an integer')
        if not isinstance(history_key, UUID):
            history_key = UUID(history_key)
        if not isinstance(since, datetime) and since is not None:
            msg = 'since must be a datetime or None'
            raise TypeError(msg)

        # Get embedding of the query string as a PGvector compatible list
        query_embedding = list(self._embedding_model.encode(text))

        # Build query
        clauses = ['main.history_key = $2\n']
        qparams = [query_embedding, history_key]
        if since is not None:
            # Don't really need the ${len(qparams) + N} thing here (first optional), but used for consistency
            clauses.append(f'ts > ${len(qparams) + 1}')
            qparams.append(since)
        if threshold is not None:
            clauses.append(THRESHOLD_WHERE_CLAUSE.format(query_threshold=f'${len(qparams) + 1}'))
            qparams.append(threshold)
        clauses = 'AND\n'.join(clauses)  # TODO: move this into the fstring below after py3.12
        where_clauses = f'WHERE\n{clauses}'
        limit_clause = f'LIMIT ${len(qparams) + 1}'
        qparams.append(limit)

        # Execute
        async with self.pool.acquire() as conn:
            message_records = await conn.fetch(
                SEMANTIC_QUERY_MESSAGE_TABLE.format(
                    table_name=self.table_name,
                    where_clauses=where_clauses,
                    limit_clause=limit_clause
                ),
                *qparams
            )

        search_results = [
            {
                'ts': record['ts'],
                'role': record['role'],
                'content': record['content'],
                'metadata': record['metadata'],
                'cosine_similarity': record['cosine_similarity']
            }
            for record in message_records
        ]

        return process_search_response(search_results)



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_llm_wrapper.py

import os
import json

import pytest

from ogbujipt import config
from ogbujipt.llm_wrapper import openai_api, openai_chat_api, prompt_to_chat # , DUMMY_MODEL
from ogbujipt.llm_wrapper import llama_cpp_http, llama_cpp_http_chat

try:
    import respx
except ImportError:
    raise RuntimeError('respx not installed. Please see wiki for setup: https://github.com/OoriData/OgbujiPT/wiki/Notes-for-contributors')  # noqa: E501

from httpx import Response

if 'OPENAI_API_KEY' in os.environ:
    del os.environ['OPENAI_API_KEY']

@pytest.mark.asyncio
@respx.mock
async def test_openai_llm_wrapper():
    # httpx_mock.add_response(
    #     url='http://127.0.0.1:8000/v1/models',
    #     json={'object':'list','data':[{'id':'model1','object':'model','owned_by':'me','permissions':[]}]})
    # httpx_mock.add_response(
    #     url='https://api.openai.com/v1/models',
    #     json={'object':'list','data':[{'id':'model1','object':'model','owned_by':'me','permissions':[]}]})

    respx.get('http://127.0.0.1:8000/v1/models').mock(return_value=Response(200))
    # FIXME: Figure out the right patterns for checking the HTTP requests
    # route1 = respx.get('http://127.0.0.1:8000/v1/models').mock(return_value=Response(200))

    host = 'http://127.0.0.1'
    api_key = 'jsbdflkajsdhfklajshdfkljalk'
    port = '8000'
    debug = True
    # model = DUMMY_MODEL
    rev = 'v1'
    base_url = f'{host}:{port}/{rev}'

    test_model = openai_chat_api(base_url=base_url, api_key=api_key, debug=debug)
    # assert route1.called

    assert test_model.api_key == api_key
    assert test_model.parameters.debug == debug
    assert test_model.model is None

    # Not OpenAI
    test_model = openai_api(base_url=base_url, debug=debug)

    assert test_model.api_key == config.OPENAI_KEY_DUMMY
    assert test_model.parameters.debug == debug
    assert test_model.model is None

    test_model = openai_chat_api(base_url=base_url, debug=debug)

    assert test_model.api_key == config.OPENAI_KEY_DUMMY
    assert test_model.parameters.debug == debug
    assert test_model.model is None
    assert test_model.base_url == base_url

    test_model = openai_chat_api(base_url=base_url, api_key=api_key, debug=debug)

    assert test_model.api_key == api_key
    assert test_model.parameters.debug == debug
    assert test_model.model is None


@pytest.mark.asyncio
@respx.mock
async def test_openai_llama_cpp_http():
    # httpx_mock.add_response(
    #     url='http://127.0.0.1:8000/v1/models',
    #     json={'object':'list','data':[{'id':'model1','object':'model','owned_by':'me','permissions':[]}]})
    # httpx_mock.add_response(
    #     url='https://api.openai.com/v1/models',
    #     json={'object':'list','data':[{'id':'model1','object':'model','owned_by':'me','permissions':[]}]})

    respx.get('http://127.0.0.1:8000/v1/models').mock(return_value=Response(200))
    route_chat = respx.post(
        'http://127.0.0.1:8000/v1/chat/completions').mock(
            return_value=Response(200, text=json.dumps(OPENAI_STYLE_CHAT_RESPONSE)))
    route_nonchat = respx.post(
        'http://127.0.0.1:8000/completion').mock(
            return_value=Response(200, text=json.dumps(OPENAI_STYLE_RESPONSE)))
    # FIXME: Figure out the right patterns for checking the HTTP requests
    # route1 = respx.get('http://127.0.0.1:8000/v1/models').mock(return_value=Response(200))

    host = 'http://127.0.0.1'
    api_key = 'jsbdflkajsdhfklajshdfkljalk'
    port = '8000'
    # model = DUMMY_MODEL
    base_url = f'{host}:{port}'

    test_endpoint = llama_cpp_http_chat(base_url=base_url, apikey=api_key)

    assert test_endpoint.apikey == api_key
    assert test_endpoint.model is None

    _ = await test_endpoint(prompt_to_chat('test'))
    assert route_chat.called

    test_endpoint = llama_cpp_http(base_url=base_url, apikey=api_key)

    assert test_endpoint.apikey == api_key
    assert test_endpoint.model is None

    _ = await test_endpoint('test')
    assert route_nonchat.called


def test_prompt_to_chat():
    chatified = prompt_to_chat('test')
    assert chatified == [{'content': 'test', 'role': 'user'}]


OPENAI_STYLE_CHAT_RESPONSE = {'choices': [{'finish_reason': 'stop',
   'index': 0,
   'message': {'content': 'Hello! How can I help you today? If you have any questions or need assistance, feel free to ask.',  # noqa: E501
    'role': 'assistant'}}],
 'created': 1710171748,
 'id': 'chatcmpl-YOGhMGqIh3qlkOzOCayFRSU3GYKY6NPJ',
 'model': 'unknown',
 'object': 'chat.completion',
 'usage': {'completion_tokens': 24, 'prompt_tokens': 11, 'total_tokens': 35},
 'prompt_tokens': 11,
 'generated_tokens': 24,
 'first_choice_text': 'Hello! How can I help you today? If you have any questions or need assistance, feel free to ask.'}  # noqa: E501

OPENAI_STYLE_RESPONSE = {'id': 'chatcmpl-Fvua40LBorK82dyfDD5QTDNl7DmeLggg',
 'choices': [{'finish_reason': 'stop',
   'index': 0,
   'message': {'content': 'Hello! How can I help you today? If you have any questions or need assistance, feel free to ask.',  # noqa: E501
    'role': 'assistant',
    'function_call': None,
    'tool_calls': None}}],
 'created': 1710172279,
 'model': 'unknown',
 'object': 'chat.completion',
 'system_fingerprint': None,
 'usage': {'completion_tokens': 24, 'prompt_tokens': 11, 'total_tokens': 35},
 'prompt_tokens': 11,
 'generated_tokens': 24,
 'first_choice_text': 'Hello! How can I help you today? If you have any questions or need assistance, feel free to ask.'}  # noqa: E501

if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")
